@@ -1,3 +1,4 @@
+# frozen_string_literal: true
 require 'rubygems/test_case'
 
 class TestGemResolver < Gem::TestCase
@@ -13,6 +14,12 @@ class TestGemResolver < Gem::TestCase
   end
 
   def set(*specs)
+    source = Gem::Source.new URI @gem_repo
+
+    specs = specs.map do |spec|
+      Gem::Resolver::SpecSpecification.new nil, spec, source
+    end
+
     StaticSet.new(specs)
   end
 
@@ -20,7 +27,7 @@ class TestGemResolver < Gem::TestCase
     actual = resolver.resolve
 
     exp = expected.sort_by { |s| s.full_name }
-    act = actual.map { |a| a.spec }.sort_by { |s| s.full_name }
+    act = actual.map { |a| a.spec.spec }.sort_by { |s| s.full_name }
 
     msg = "Set of gems was not the same: #{exp.map { |x| x.full_name}.inspect} != #{act.map { |x| x.full_name}.inspect}"
 
@@ -89,24 +96,6 @@ class TestGemResolver < Gem::TestCase
     assert_same index_set, composed
   end
 
-  def test_handle_conflict
-    a1 = util_spec 'a', 1
-
-    r1 = Gem::Resolver::DependencyRequest.new dep('a', '= 1'), nil
-    r2 = Gem::Resolver::DependencyRequest.new dep('a', '= 2'), nil
-    r3 = Gem::Resolver::DependencyRequest.new dep('a', '= 3'), nil
-
-    existing = Gem::Resolver::ActivationRequest.new a1, r1, false
-
-    res = Gem::Resolver.new [a1]
-
-    res.handle_conflict r2, existing
-    res.handle_conflict r2, existing
-    res.handle_conflict r3, existing
-
-    assert_equal 2, res.conflicts.length
-  end
-
   def test_requests
     a1 = util_spec 'a', 1, 'b' => 2
 
@@ -116,11 +105,35 @@ class TestGemResolver < Gem::TestCase
 
     res = Gem::Resolver.new [a1]
 
-    reqs = Gem::Resolver::RequirementList.new
+    reqs = []
 
     res.requests a1, act, reqs
 
-    assert_equal ['b (= 2)'], reqs.to_a.map { |req| req.to_s }
+    assert_equal ['b (= 2)'], reqs.map { |req| req.to_s }
+  end
+
+  def test_requests_development
+    a1 = util_spec 'a', 1, 'b' => 2
+
+    spec = Gem::Resolver::SpecSpecification.new nil, a1
+    def spec.fetch_development_dependencies
+      @called = true
+    end
+
+    r1 = Gem::Resolver::DependencyRequest.new dep('a', '= 1'), nil
+
+    act = Gem::Resolver::ActivationRequest.new spec, r1, false
+
+    res = Gem::Resolver.new [act]
+    res.development = true
+
+    reqs = []
+
+    res.requests spec, act, reqs
+
+    assert_equal ['b (= 2)'], reqs.map { |req| req.to_s }
+
+    assert spec.instance_variable_defined? :@called
   end
 
   def test_requests_ignore_dependencies
@@ -133,13 +146,111 @@ class TestGemResolver < Gem::TestCase
     res = Gem::Resolver.new [a1]
     res.ignore_dependencies = true
 
-    reqs = Gem::Resolver::RequirementList.new
+    reqs = []
 
     res.requests a1, act, reqs
 
     assert_empty reqs
   end
 
+  def test_resolve_conservative
+    a1_spec = util_spec 'a', 1
+    a2_spec = util_spec 'a', 2 do |s|
+      s.add_dependency 'b', 2
+      s.add_dependency 'c'
+    end
+    b1_spec = util_spec 'b', 1
+    b2_spec = util_spec 'b', 2
+    c1_spec = util_spec 'c', 1 do |s| s.add_dependency 'd', 2 end
+    c2_spec = util_spec 'c', 2 do |s| s.add_dependency 'd', 2 end
+    d1_spec = util_spec 'd', 1 do |s| s.add_dependency 'e' end
+    d2_spec = util_spec 'd', 2 do |s| s.add_dependency 'e' end
+    e1_spec = util_spec 'e', 1
+    e2_spec = util_spec 'e', 2
+
+    a_dep = make_dep 'a', '= 2'
+    e_dep = make_dep 'e'
+
+    # When requesting to install:
+    # a-2, e
+    deps = [a_dep, e_dep]
+
+    s = set a1_spec, a2_spec, b1_spec, b2_spec, c1_spec, c2_spec, d1_spec, d2_spec, e1_spec, e2_spec
+
+    res = Gem::Resolver.new deps, s
+
+    # With the following gems already installed:
+    # a-1, b-1, c-1, e-1
+    res.skip_gems = {'a'=>[a1_spec], 'b'=>[b1_spec], 'c'=>[c1_spec], 'e'=>[e1_spec]}
+
+    # Make sure the following gems end up getting used/installed/upgraded:
+    # a-2 (upgraded)
+    # b-2 (upgraded), specific dependency from a-2
+    # c-1 (used, not upgraded), open dependency from a-2
+    # d-2 (installed), specific dependency from c-2
+    # e-1 (used, not upgraded), open dependency from request
+    assert_resolves_to [a2_spec, b2_spec, c1_spec, d2_spec, e1_spec], res
+  end
+
+  def test_resolve_development
+    a_spec = util_spec 'a', 1 do |s| s.add_development_dependency 'b' end
+    b_spec = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    c_spec = util_spec 'c', 1
+
+    a_dep = make_dep 'a', '= 1'
+
+    deps = [a_dep]
+
+    s = set a_spec, b_spec, c_spec
+
+    res = Gem::Resolver.new deps, s
+
+    res.development = true
+
+    assert_resolves_to [a_spec, b_spec, c_spec], res
+  end
+
+  def test_resolve_development_shallow
+    a_spec = util_spec 'a', 1 do |s|
+      s.add_development_dependency 'b'
+      s.add_runtime_dependency 'd'
+    end
+
+    b_spec = util_spec 'b', 1 do |s| s.add_development_dependency 'c' end
+    c_spec = util_spec 'c', 1
+
+    d_spec = util_spec 'd', 1 do |s| s.add_development_dependency 'e' end
+    e_spec = util_spec 'e', 1
+
+    a_dep = make_dep 'a', '= 1'
+
+    deps = [a_dep]
+
+    s = set a_spec, b_spec, c_spec, d_spec, e_spec
+
+    res = Gem::Resolver.new deps, s
+
+    res.development = true
+    res.development_shallow = true
+
+    assert_resolves_to [a_spec, b_spec, d_spec], res
+  end
+
+  def test_resolve_remote_missing_dependency
+    @fetcher = Gem::FakeFetcher.new
+    Gem::RemoteFetcher.fetcher = @fetcher
+
+    a_dep = make_dep 'a', '= 1'
+
+    res = Gem::Resolver.new [a_dep], Gem::Resolver::IndexSet.new
+
+    e = assert_raises Gem::UnsatisfiableDepedencyError do
+      res.resolve
+    end
+
+    refute_empty e.errors
+  end
+
   def test_no_overlap_specificly
     a = util_spec "a", '1'
     b = util_spec "b", "1"
@@ -215,7 +326,7 @@ class TestGemResolver < Gem::TestCase
 
     res = Gem::Resolver.new([ad], s)
 
-    assert_resolves_to [a2_p1], res
+    assert_resolves_to [a2_p1.spec], res
   end
 
   def test_only_returns_spec_once
@@ -310,19 +421,19 @@ class TestGemResolver < Gem::TestCase
       r.resolve
     end
 
-    deps = [make_dep("c", "= 1"), make_dep("c", "= 2")]
+    deps = [make_dep("c", "= 2"), make_dep("c", "= 1")]
     assert_equal deps, e.conflicting_dependencies
 
     con = e.conflict
 
     act = con.activated
-    assert_equal "c-2", act.spec.full_name
+    assert_equal "c-1", act.spec.full_name
 
     parent = act.parent
-    assert_equal "b-1", parent.spec.full_name
+    assert_equal "a-1", parent.spec.full_name
 
     act = con.requester
-    assert_equal "a-1", act.spec.full_name
+    assert_equal "b-1", act.spec.full_name
   end
 
   def test_raises_when_a_gem_is_missing
@@ -406,15 +517,15 @@ class TestGemResolver < Gem::TestCase
 
     dependency = e.conflict.dependency
 
-    assert_equal 'a', dependency.name
+    assert_includes %w(a b), dependency.name
     assert_equal req('>= 0'), dependency.requirement
 
     activated = e.conflict.activated
-    assert_equal 'c-1', activated.full_name
+    assert_equal 'c-2', activated.full_name
 
-    assert_equal dep('c', '= 1'), activated.request.dependency
+    assert_equal dep('c', '>= 2'), activated.request.dependency
 
-    assert_equal [dep('c', '>= 2'), dep('c', '= 1')],
+    assert_equal [dep('c', '= 1'), dep('c', '>= 2')],
                  e.conflict.conflicting_dependencies
   end
 
@@ -602,4 +713,3 @@ class TestGemResolver < Gem::TestCase
   end
 
 end
-
